#ifndef BisqNES_included
#define BisqNES_included

#include "cpu.hh"
#include "ppu.hh"
#include "papu.hh"
#include "pads.hh"
#include "clock.h"

#include "mappers.hh" // FIRST INCLUSION

class bisqnes
{
private:
    friend mapper;
    friend bisqpads;
    
    m6502cpu cpu;
    
    bisqppu ppu;
    bisqpapu papu;
    bisqpads pads;
    
    BYTE ram[0x800];   // Mirrored four times
    BYTE sram[0x2000]; // Battery RAM
    
    BYTE *rom;
    BYTE rom16count;
    BYTE ctrlbyte;
    BYTE vrom8count;
    BYTE mappernum;
    
    BYTE *rompage[4]; // 8k pages (0x2000)
    
    mapper *mapp;
    
public:
    bisqnes() : rom(NULL), mapp(NULL)
    {
    }
    ~bisqnes()
    {
        if(rom)delete[] rom;
    }
    
    void load(const char *fn);
    
    void poweron()
    {
        memset(ram, 0, sizeof ram);
        memset(sram, 0, sizeof sram);
        
        mapp->reset();
        mapp->initialization();
        cpu.Reset();
    }
    
    void sta_load(const char *fn);
    void snss_load(const char *fn)
    {
    	FILE *fp = fopen(fn, "rb");
    	if(!fp)return;
    	
    	if(fgetc(fp)!='S' || fgetc(fp)!='N' || fgetc(fp)!='S' || fgetc(fp)!='S')
    	{
    		fclose(fp);
    		return;    	
    	}
    	
    	fclose(fp);
        fprintf(stderr, "\nLoaded %s.\n", fn);
    }
public:
    inline void Wr(register fWORD Addr, register fBYTE Value)
    {
        if(Addr < 0x2000) { ram[Addr & 0x7FF] = Value; return; }
        if(Addr < 0x6000)
        {
            if(Addr == 0x4014)
            {
                /* DMA transfer source address */
                Addr = Value * 0x100;
                /* This _should_ move one byte per 6 tick()s,
                 * but it's hard to write.
                 */
                for(unsigned a=0; a<256; ++a)
                    ppu.spr_ram[a] = Rd(Addr++);
                return;
            }
            if(ppu.write(Addr, Value))return;
            if(papu.write(Addr, Value))return;
            if(pads.write(Addr, Value))return;
        }
        else if(Addr < 0x8000)
        {
            if(ctrlbyte&8)
                sram[Addr & 0x1FFF] = Value;
        }
        else
            mapp->write(Addr, Value);
    }
    inline BYTE Rd(register fWORD Addr)
    {
        if(Addr >= 0x8000)return rompage[(Addr>>13) & 3][Addr & 0x1FFF];
        if(Addr < 0x2000)return ram[Addr & 0x7FF];
        if(Addr < 0x6000)
        {
            BYTE value = 0;
            if(ppu.read(Addr, value))return value;
            if(papu.read(Addr, value))return value;
            if(pads.read(Addr, value))return value;
        }
        return sram[Addr & 0x1FFF];
    }
    inline BYTE Op(register fWORD Addr) const
    {
        if(Addr & 0x8000)return rompage[(Addr>>13) & 3][Addr & 0x1FFF];
        if(Addr & 0x4000)return sram[Addr & 0x1FFF];
        return ram[Addr & 0x7FF];
    }
    void run();
};

extern bisqnes nes;

#endif
